/*
  asm.cc du cours ProjectCompil
  Ce fichier est distrubué sous licence GPL
*/
#include <cstdlib>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <string>
#include <list>
#include <map>
#include <vector>
#include <math.h>
#include <locale>

using namespace std;

int verbose;
bool pseudoinstruction;

vector<uint16_t> hexcode;      // the code is built into this vector
map<string, uint16_t> label;   // the labels 

// A structure set up in the first pass, and used in the second pass to resolve labels
enum atr_type {JMI, JMRI, JMSI, LDL };

struct address_to_resolve {
  uint16_t hexcode_index; // index in the hexcode vector. This is the output code vector
  atr_type type;          // This is the instruction type
  string label;           // This is the label to be resolved
};

vector<address_to_resolve> todo; //this is a vector of labels which were not resolved during
                                 //the first pass

uint16_t line_number;     // the line number in the input .asm file
uint16_t base_address = 0x0800; // address at which the code will be compiled
uint16_t constant_address = 0x1FFF; // address at which the constants will be placed. 
                                    // this is the last ROM address. Other constants
                                    // are placed at smaller ROM addresses
uint16_t variable_address = 0x07FF; //just before the RAM;

map<string, uint16_t> c_addr; //the constants -> map constant name to memory address
map<string, uint16_t> c_val; //the constants -> map constant name to initialization values //may be useless
map<uint16_t, uint16_t> c_initVal; //the constants -> map constant address to the init values

map<string, uint16_t> v_name_addr; //the variables -> map variable name to memory address
map<uint16_t, string>   v_addr_name;   //the variables -> map the memory address to the variable names

uint16_t current_address; // the current address in the code section of the program

//------------------------------------------------------------------------------
/** Return the code of the condition predicate. 
 * @param &slist - the input token list
 * @return - the code of the condition predicate. If no condition predicate 
 * exist, the return code is 7. This the code for NC - no condition. 
 */
uint16_t read_condpred(list<string> &slist) {
	int pred_code=0;
	string s = slist.front();     // fetch the next token from sList
	if (string(s, 0, 2) != "?p")  //implicit predicate p8 = "no condition"
		return 7;
	else {
		slist.pop_front(); 
		s = slist.front(); //Now the options are: GT, GE, EQ, NE, LE, OV, NC
		if      (s == "GT") pred_code = 0;
		else if (s == "GE") pred_code = 1;
		else if (s == "EQ")	pred_code = 2;
		else if (s == "NE") pred_code = 3;
		else if (s == "LE")	pred_code = 4;
		else if (s == "LT") pred_code = 5;
		else if (s == "OV")	pred_code = 6;
		else if (s == "NC")	pred_code = 7;
		else{ 
			std::ostringstream o;
			o << "Error in " <<  __FILE__ << "@" << __LINE__ << ". Invalid option for predicate: got " << s << " expected: "
			  << "one of GT, GE, EQ, NE, LE, LT, OV, NC.";
			throw o.str();
		}
		slist.pop_front();
	}
	return pred_code;
}

//------------------------------------------------------------------------------
/** Return the value of the s bit in the instruction 
 * @param &slist - the input token list
 * @return - value of the s bit. This dictates it the current instruction 
 * writes or not the FLAG register. 
 */
uint16_t read_sr(list<string> &slist) {
  string s = slist.front();
  if (string(s, 0, 2).compare("!s")==0)  {
	cerr << "FOUND S=1"<<endl;  
	slist.pop_front();
	return 1;	
  }else
	return 0;
}

//------------------------------------------------------------------------------
/** Returns the register number 
 * @param &slist - the input token list
 * @return - the index of the selected register. For example, for register 
 * r42, the function will return 42  
 */
uint16_t read_reg(list<string> &slist) {
  string s = slist.front();
  if (s[0] != 'r')  {
    cerr << "Error, line " << line_number << ": expected register, got " << s << "'." << endl;
    exit(-1);
  }else{
    s.erase(0, 1);
	int val;    
	istringstream valstr(s);
    valstr >> val;
    if (val < 0 || val > 127) {
      cerr << "Error, line " << line_number << ": invalid register r" << s << "'." << endl;
      exit(-1);
    }
    slist.pop_front();
    return val;
  }
}

//------------------------------------------------------------------------------
/** Returns the variable 
 * @param &slist - the input token list
 * @return - the variable name
 */
string read_var_name(list<string> &slist) {
  string s = slist.front(); //the variable's name
  slist.pop_front();
  return s;
}


//------------------------------------------------------------------------------
/** Return an integer value 
 * @param s - the string to be converted in an integer. The string could contain
 * the integer in base 10, or in hexadecimal, preceeded by the 0x string.
 * @return - the integer value of the string.
 */
int read_int(string s) {
  int val;
  if(string(s, 0, 2) == "0x") {
    s.erase(0, 2);
    istringstream valstr(s);
    valstr >> hex >> val;
  }
  else {
    istringstream valstr(s);
    valstr >> val;
  }
  return val;
}

//------------------------------------------------------------------------------
/** Return the immediate value 
 * @param &slist - the input token list 
 * @return - the integer 
 */
uint8_t read_imm8(list<string>  & slist)
{
  string s = slist.front();
  int val = read_int(s);
  if(val<0)
    val+=256;
  if (val < 0 || val > 255) {
    cerr << "Error, line " << line_number << ": invalid 8-bit immediate constant " << s << "'." << endl;
    exit(-1);
  }
  slist.pop_front();
  return val;
}

//------------------------------------------------------------------------------
/** Return the immediate value 
 * @param &slist - the input token list 
 * @return - the integer 
 */
uint8_t read_imm16(list<string>  & slist){
  string s = slist.front();
  uint8_t val = read_int(s);
  if (val < 0 || val > 65535) {
    cerr << "Error, line " << line_number << ": invalid 16-bit immediate constant " << s << "'." << endl;
    exit(-1);
  }
  slist.pop_front();
  return val;
}


//------------------------------------------------------------------------------
/** Return the immediate value, if available. If the decimal/hex integer is
 * replaced by a label, the label is inserted into a todo list (list of labels
 * which are processed during the second pass. 
 * @param &slist - the input token list
 * @param atrt - the instruction type from which the function was called 
 * @return - 0 is label is found/ the correspoding integer value otherwise 
 */
uint16_t read_imm8_or_label(list<string>  & slist, atr_type atrt){
  string s = slist.front();
  if(s[0]<'0' || s[0]>'9') { //assume a label
    address_to_resolve atr;
    atr.hexcode_index = hexcode.size();
    atr.type = atrt;
    atr.label = s;
    todo.push_back(atr);
    slist.pop_front();
    return 0;
  }
  else { // assume a numeric constant
    uint16_t val = read_int(s);
    if (val < 0 || val > 255) {
      cerr << "Error, line " << line_number << ": invalid 8-bit immediate constant " << s << "'." << endl;
      exit(-1);
    }
    slist.pop_front();
    return val;
  }
}

//------------------------------------------------------------------------------
uint16_t ldl(list<string>  &slist, uint16_t p){
	uint16_t c = 0;
    slist.pop_front();
    c += 0 <<11; //opcode is 00000
	c += read_imm8(slist);
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t ldl(uint16_t val){
    uint16_t c = 0;
    c += 0 <<11; //opcode is 00000
	c += 7 << 8; //predicate = 7	
	c += val;
	return c;
}

//------------------------------------------------------------------------------
uint16_t ldh(list<string>  &slist, uint16_t p){
	uint16_t c = 0;    
	slist.pop_front();
    c += 1<<11;
    c += read_imm8(slist);
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t ldh(uint16_t val){
    uint16_t c = 0;
    c += 1 <<11; //opcode is 00001
	c += 7 << 8; //predicate = 7	
	c += val ;
	return c;
}
//------------------------------------------------------------------------------
uint16_t mv2reg(list<string>  &slist, uint16_t p){
	uint32_t c=0;    
	slist.pop_front();
    c += 2<<11;
    c += read_reg(slist);
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t mv2acc(list<string>  &slist, uint16_t p){
	uint32_t c=0;    
    slist.pop_front();
    c += 3 << 11;
    c += read_reg(slist);
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t mv2acc(uint8_t r, uint8_t p){
	uint32_t c=0;    
    c += 3 << 11;
    c += r;
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t rd(list<string>  &slist, uint16_t p){
	uint32_t c=0;    
    slist.pop_front();
    c += 4 << 11;
    c += read_reg(slist);
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t rd(uint16_t reg_number){
	uint32_t c=0;    
    c += 4 << 11;
    c += reg_number;
	c += 7 << 8;
	return c;
}


//------------------------------------------------------------------------------
uint16_t wr(list<string>  &slist, uint16_t p){
	uint32_t c=0;    
    slist.pop_front();
    c += 5 << 11;
    c += read_reg(slist);
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t rdinc(list<string>  &slist, uint16_t p){
	uint32_t c=0;
    slist.pop_front();
    c += 6 << 11;
    c += read_reg(slist);
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t wrinc(list<string>  &slist, uint16_t p){
	uint32_t c=0;
    slist.pop_front();
    c += 7 << 11;
    c += read_reg(slist);
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t rddec(list<string>  &slist, uint16_t p){
	uint32_t c=0;
    slist.pop_front();
    c += 8 << 11;
    c += read_reg(slist);
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t wrdec(list<string>  &slist, uint16_t p){
	uint32_t c=0;
    slist.pop_front();
    c += 9 << 11;
    c += read_reg(slist);
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t jmp(list<string>  &slist, uint16_t p){
	uint32_t c=0;
    slist.pop_front();
    c += 10 << 11;
    c += read_reg(slist);
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t jmr(list<string>  &slist, uint16_t p){
	uint32_t c=0;
    slist.pop_front();
    c += 11 << 11;
    c += read_reg(slist);
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t jmi(list<string>  &slist, uint16_t p){
	uint32_t c=0;
    slist.pop_front();
    c += 12 << 11;
    c += read_imm8_or_label(slist,JMI);
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t jmri(list<string>  &slist, uint16_t p){
	uint32_t c=0;
    slist.pop_front();
    c += 13 << 11;
    c += read_imm8_or_label(slist,JMRI);
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t jsr(list<string>  &slist, uint16_t p){
	uint32_t c=0;
    slist.pop_front();
    c += 14 << 11;
    c += read_reg(slist);
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t jsri(list<string>  &slist, uint16_t p){
	uint32_t c=0;
    slist.pop_front();
    c += 15 << 11;
    c += read_imm8_or_label(slist,JMSI);
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t rts(list<string>  &slist, uint16_t p){
	uint32_t c=0;
    slist.pop_front();
    c += 16 << 11;
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t nop(list<string>  &slist, uint16_t p){
	uint32_t c=0;
    slist.pop_front();
    c += 17 << 11;
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t add(list<string>  &slist, uint16_t p, uint16_t s){
	uint32_t c=0;
    slist.pop_front();
    c += 18 << 11;
    c += read_reg(slist);
    c += s << 7;	
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t sub(list<string>  &slist, uint16_t p, uint16_t s){
	uint32_t c=0;
    slist.pop_front();
    c += 19 << 11;
    c += read_reg(slist);
    c += s << 7;	
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t sub(uint8_t r, uint8_t p, uint8_t s){
	uint32_t c = 0;
    c += 19 << 11;
    c += r;
    c += s << 7;
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t mul(list<string>  &slist, uint16_t p, uint16_t s){
	uint32_t c=0;
    slist.pop_front();
    c += 20 << 11;
    c += read_reg(slist);
    c += s << 7;
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t cmp(list<string>  &slist, uint16_t p, uint16_t s){
	uint32_t c=0;
    slist.pop_front();
    c += 21 << 11;
    c += read_reg(slist);
    c += s << 7;
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t cmp(uint8_t r, uint8_t p, uint8_t s){
	uint32_t c = 0;
    c += 21 << 11;
    c += r;
    c += s << 7;
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t swap(list<string>  &slist, uint16_t p, uint16_t s){
	uint32_t c=0;
    slist.pop_front();
    c += 22 << 11;
    c += read_reg(slist);
    c += s << 7;
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t clr(list<string>  &slist, uint16_t p, uint16_t s){
	uint32_t c=0;
    slist.pop_front();
    c += 23 << 11;
    c += read_reg(slist);
    c += s << 7;
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t and_i(list<string>  &slist, uint16_t p, uint16_t s){
	uint32_t c=0;
    slist.pop_front();
    c += 24 << 11;
    c += read_reg(slist);
    c += s << 7;
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t or_i(list<string>  &slist, uint16_t p, uint16_t s){
	uint32_t c=0;
    slist.pop_front();
    c += 25 << 11;
    c += read_reg(slist);
    c += s << 7;
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t xor_i(list<string>  &slist, uint16_t p, uint16_t s){
	uint32_t c=0;
    slist.pop_front();
    c += 26 << 11;
    c += read_reg(slist);
    c += s << 7;
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t not_i(list<string>  &slist, uint16_t p, uint16_t s){
	uint32_t c=0;
    slist.pop_front();
    c += 27 << 11;
    c += read_reg(slist);
    c += s << 7;
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t lsr(list<string>  &slist, uint16_t p, uint16_t s){
	uint32_t c=0;
    slist.pop_front();
    c += 28 << 11;
    c += read_reg(slist);
    c += s << 7;
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t lsl(list<string>  &slist, uint16_t p, uint16_t s){
	uint32_t c=0;
    slist.pop_front();
    c += 29 << 11;
    c += read_reg(slist);
    c += s << 7;
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t ror(list<string>  &slist, uint16_t p, uint16_t s){
	uint32_t c=0;
    slist.pop_front();
    c += 30 << 11;
    c += read_reg(slist);
    c += s << 7;
	c += p << 8;
	return c;
}
//------------------------------------------------------------------------------
uint16_t rol(list<string>  &slist, uint16_t p, uint16_t s){
	uint32_t c=0;
    slist.pop_front();
    c += 31 << 11;
    c += read_reg(slist);
    c += s << 7;
	c += p << 8;
	return c;
}


/** Function which return s the machine code of the next instruction in slist
 * @param &slist - the input token list
 * @return - the machine code of the instruction
 */
uint16_t read_op(list<string>  & slist) {

  pseudoinstruction = false;
  uint16_t code = 0;
  uint16_t pred;
  uint16_t sr = read_sr(slist);
	  
  
  // read the predicates if they exist, and put its number in place
  try{
  	pred = read_condpred(slist);
  }catch(std::string s){
		cerr << "Exception while parsing command line: " << s << endl;
		exit(-1);
  }
  // now read the instruction itself
  
  locale loc;
  for(size_t i=0;i<slist.front().length();i++)
    slist.front()[i]=tolower(slist.front()[i],loc);

  if      (slist.front() == "ldl"   ) { code = ldl   (slist, pred); }
  else if (slist.front() == "ldh"   ) { code = ldh   (slist, pred); }
  else if (slist.front() == "mv2reg") { code = mv2reg(slist, pred); }
  else if (slist.front() == "mv2acc") { code = mv2acc(slist, pred); }
  else if (slist.front() == "rd"    ) { code = rd    (slist, pred); }
  else if (slist.front() == "wr"    ) { code = wr    (slist, pred); }
  else if (slist.front() == "rdinc" ) { code = rdinc (slist, pred); }
  else if (slist.front() == "wrinc" ) { code = wrinc (slist, pred); } 
  else if (slist.front() == "rddec" ) { code = rddec (slist, pred); }
  else if (slist.front() == "wrdec" ) { code = wrdec (slist, pred); }
  else if (slist.front() == "jmp"   ) { code = jmp   (slist, pred); }
  else if (slist.front() == "jmr"   ) { code = jmr   (slist, pred); }
  else if (slist.front() == "jmi"   ) { code = jmi   (slist, pred); }
  else if (slist.front() == "jmri"  ) { code = jmri  (slist, pred); }
  else if (slist.front() == "jsr"   ) { code = jsr   (slist, pred); }
  else if (slist.front() == "jsri"   ) { code = jsri   (slist, pred); }
  else if (slist.front() == "rts"   ) { code = rts   (slist, pred); }
  else if (slist.front() == "nop"   ) { code = nop   (slist, pred); }
  else if (slist.front() == "add"   ) { code = add   (slist, pred, sr); }
  else if (slist.front() == "sub"   ) { code = sub   (slist, pred, sr); }
  else if (slist.front() == "mul"   ) { code = mul   (slist, pred, sr); }
  else if (slist.front() == "cmp"   ) { code = cmp   (slist, pred, sr); }
  else if (slist.front() == "swap"  ) { code = swap  (slist, pred, sr); }
  else if (slist.front() == "clr"   ) { code = clr   (slist, pred, sr); }
  else if (slist.front() == "and"   ) { code = and_i (slist, pred, sr); }
  else if (slist.front() == "or"    ) { code = or_i  (slist, pred, sr); }
  else if (slist.front() == "xor"   ) { code = xor_i (slist, pred, sr); }
  else if (slist.front() == "not"   ) { code = not_i (slist, pred, sr); }
  else if (slist.front() == "lsr"   ) { code = lsr   (slist, pred, sr); }
  else if (slist.front() == "lsl"   ) { code = lsl   (slist, pred, sr); }
  else if (slist.front() == "ror"   ) { code = ror   (slist, pred, sr); }
  else if (slist.front() == "rol"   ) { code = rol   (slist, pred, sr); }
//------------------------------------------------------------------------------
// PSEUDOINSTRUCTIONS
  else if (slist.front() == "abs") {
	pseudoinstruction = true;
    slist.pop_front();
	//----------------
    uint16_t code = ldl(0); //load 0 to acc
   	hexcode.push_back(code);//put the code into hexcode
	current_address ++;
	//----------------
	uint8_t reg = read_reg(slist);
    code = sub(reg,7,1); //compare and set flags
   	hexcode.push_back(code);//put the code into hexcode
	current_address ++;
	//----------------
	code = mv2acc(reg,5); //conditional mv2acc
   	hexcode.push_back(code);//put the code into hexcode
	current_address ++;
	//----------------
  } else if (slist.front() == "block") {
		//an array behaves like any variable, except that it has a variable size allocation
		pseudoinstruction = true;
		slist.pop_front();		
		string var = read_var_name(slist);
		cerr << "Array found with the name: " << var << endl;
		uint8_t arraySize = read_imm8(slist);	//TODO Check that array size is not negative	
		v_name_addr[var] = variable_address;
		v_addr_name[variable_address] = var;
		variable_address=variable_address - arraySize;  
  } else if (slist.front() == "integer") {
		pseudoinstruction = true;
		slist.pop_front();		
		string var = read_var_name(slist);
		cerr << "Variable found with the name: " << var << endl;
		v_name_addr[var] = variable_address;
		v_addr_name[variable_address] = var;
		variable_address--;  
  } else if (slist.front() == "define") { //constants should be initializad at definition
		pseudoinstruction = true;
		slist.pop_front();		
		string var = read_var_name(slist);
		cerr << "Constant found with the name: " << var << endl;
		uint16_t initValue = read_imm16(slist); 
		cerr << "Constant " << var << " has value: "<< initValue<< endl;		
		//if everything went well so far, 
		c_addr[var] = constant_address;
		c_val[var] = initValue;
        c_initVal[constant_address] = initValue;
		constant_address--;
  } else if (slist.front() == "ldvar") {
		//loads a variable into a register
		pseudoinstruction = true;
		slist.pop_front();		
		string var = read_var_name(slist);
		cerr << "Variable found with the name: " << var << endl;
		uint8_t rx = read_reg(slist);
		uint16_t var_addr = v_name_addr[var];
		uint16_t low_address = 0x00FF & var_addr;
		uint16_t high_address = var_addr >> 8;		

	    uint16_t code = ldl( low_address );
	   	hexcode.push_back(code);//put the code into hexcode
		current_address ++;
				
	    code = ldh( high_address );
	   	hexcode.push_back(code);//put the code into hexcode
		current_address ++;

	    code = rd( rx );
	   	hexcode.push_back(code);//put the code into hexcode
		current_address ++;
	} else if (slist.front() == "loadl") {
		pseudoinstruction = true;
		slist.pop_front();		
		string var = read_var_name(slist);
		//break the varname'direction into varname and direction
		string varname = var.substr(0, var.find("'"));
		string dir = var.substr(var.find("'")+1);
		//check if the variable was defined
		map<string,uint16_t>::iterator iter = v_name_addr.find(varname);
		if (iter == v_name_addr.end()){ 
			cerr << "Ooooooops.... variable "<<varname <<" was not declared"<<endl;
			exit(1);
		}
		//check if the direction is correct
		if (!  (dir.compare("high")==0 || dir.compare("low")==0)){
			cerr << "Oopsss ....... Bad Direction" <<endl;
			exit(1);
		}
		uint16_t value;
		if (!dir.compare("high")) //comapre function return 0 if strings match
			value = v_name_addr[varname] >> 8;
		else
			value = (0x00FF & v_name_addr[varname]);

		code = ldl(value);
	   	hexcode.push_back(code);//put the code into hexcode
		current_address ++;
		
		
   } else if (slist.front() == "loadh") {
		pseudoinstruction = true;
		slist.pop_front();		
		string var = read_var_name(slist);
		//break the varname'direction into varname and direction
		string varname = var.substr(0, var.find("'"));
		string dir = var.substr(var.find("'")+1);
		//check if the variable was defined
		map<string,uint16_t>::iterator iter = v_name_addr.find(varname);
		if (iter == v_name_addr.end()){ 
			cerr << "Ooooooops.... variable "<<varname <<" was not declared"<<endl;
			exit(1);
		}
		//check if the direction is correct
		if (!(!dir.compare("high") || !dir.compare("low"))){
			cerr << "Oopsss ....... Bad Direction" <<endl;
			exit(1);
		}

		uint16_t value;
		if (!dir.compare("high")) //comapre function return 0 if strings match
			value = v_name_addr[varname] >> 8;
		else
			value = (0x00FF & v_name_addr[varname]);

		code = ldh(value);
	   	hexcode.push_back(code);//put the code into hexcode
		current_address ++;
   }
   else {
        cerr<<"Instruction '"<<slist.front()<<"' does not exist !\n";
        exit(-1);
   }

  return code;
}


//------------------------------------------------------------------------------
int main(int argc, char* argv[]){

  verbose = 0;
  if (argc>1) {
    string opt(argv[1]);
    if(opt=="-v")
      verbose=1;
  }

  line_number     = 0;            // the line in the source code
  current_address = base_address; // the address in the object code
  
  // Pass 1 
  if(verbose){
    cerr << "**** PASS 1 ****" << endl;
  }
  while (!cin.eof()) {

    // read a line
    char buf[256];
    cin.getline(buf, 256);
    string line(buf);
    line_number++;
    if (line.empty())
      continue;
    
    // break the line into tokens
    istringstream sstr(line);
    list<string> slist;
    while (!sstr.eof()) {
      string str;
      sstr >> str;
      if (str[0] == ';') // comment: stop parsing here
        break;      
      else{
        if (!str.empty()) 
          slist.push_back(str);
      }
    }

    if (!slist.empty()) {     // Now process the tokens
      string s;
      uint16_t code = 0;

      s = slist.front();

      //PROCESS LABEL
	  //========================================================================
      // is the token a label ?
      if (s[s.length()-1] == ':') {
        string id = s.substr(0, s.length()-1);
        // Does the label already exist ?
        if (label.find(id) != label.end()) {
          cerr << "Error, label " << id  << " already defined " << endl;
          exit(-1);  
        }
        label[id] = current_address;        
        slist.pop_front(); // remove it
        if(verbose){
          cerr << "label " << id << " @ " << setfill('0')  << setw(4) << hex << current_address <<endl;
        }
      }
      //========================================================================
	
      if (!slist.empty()) {     //proceed, if this is not a label
        s = slist.front();
        
		//normal instructions
        uint16_t instr = read_op(slist); //get the instruction code
		if (not pseudoinstruction){
        	hexcode.push_back(instr);//put the code into hexcode
			current_address ++;
		}else{
			//for the pseudoinstructions, the hexcode is pushed back directly from read_op, and 
        	//current_address is incremented accordingly there        
		}

		if(verbose){
        	cerr << "@ " << setfill('0')  << setw(4) << hex << current_address-1  //FIXME verbose printings
                << ": "  << setw(4) << hex << code << endl;
        }
        
      }
    }
  }
  // Pass 2
if(verbose){
  cerr << "**** PASS 2 ****" << endl;
}

  // Fix all the missing labels
  for (uint16_t i=0; i<todo.size(); i++) {
    address_to_resolve atr = todo[i];
    uint16_t index = atr.hexcode_index;
    atr_type type = atr.type;
    string s = atr.label;
    uint16_t instr;

    // look up the label
    if (label.find(s) == label.end()) {
          cerr << "Error, label " << s  << " is used but not defined " << endl;
          exit(-1);  
    }
    // OK, the label exists
    uint16_t resolved_address = label[s];      
    uint16_t instr_address = base_address + index;
    // Now put the address to its proper place
    if ((type==JMRI) || (type==JMSI)) {
      instr = hexcode[index];
      int delta = resolved_address - instr_address;
      if(delta<-127 || delta>127) {
          cerr << "Error, jmri to label " << s  << " out of range " << endl;
          exit(-1);  
        }
      
	  if (delta<0)
		delta = 256 + delta; //2's complement
      instr=instr+(delta);
      hexcode[index] = instr;
      if(verbose){
        cerr << "Resolved label " << s 
             << ", at address " << setfill('0')  << setw(4) << hex << instr_address
             << " code is now " << setfill('0')  << setw(4) << hex << instr << endl;
      }
    }
    
  }


  uint16_t code_size = hexcode.size();
  uint16_t constant_size = 0x2000 - constant_address;
  
  if (code_size + constant_size > 0x1800 )
	cerr << "Classical error - > does not fit" <<endl;
  else{
	  
	  // Output
	  if(verbose){
		cerr << "Outputing object code..." << endl;
	  }
	  for (uint16_t i=0; i<hexcode.size(); i++) {
		  cout << setfill('0')  << setw(4) << hex << hexcode[i] << endl;
	  }
	  // Pad with zeroes
	  for (uint16_t i=0; i< (0x1800 - (code_size + constant_size)); i++) {
		  cout << setfill('0')  << setw(4) << hex << 0 << endl;
	  }
	  // Place constants
	  for (uint16_t i=0x2000-constant_size ; i< 0x2000; i++) {
		  cout << setfill('0')  << setw(4) << hex << c_initVal[i] << endl;
	  }

	  if(verbose){
		cerr << "...done" << endl;
	  }

  }
}


