/*
  simu.cc du cours ASR1 2008
  Ce fichier est distrubué sous licence GPL
*/
#include <iostream>
#include <iomanip>
#include <sstream>
#include <fstream>
#include <string>
#include <list>
#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <cstdlib>
#include <math.h>
#include <stdexcept>
#include <ncurses.h>
#include <map>
#include <signal.h>

using namespace std;

#include "../common/disasm.hpp"

#define NREGS 128

struct coord {
  uint16_t row; 
  uint16_t col; 
};

int verbose;
int step_by_step;

#define TOPSCREEN 0x5000
#define TOPRAM 0x8000

int SKIP_FRAME=100;

// SIMULATOR DEPENDENT
uint32_t video[640*480]; //the size of the video memory


//PROCESSOR DEPENDENT
uint16_t r[128]; //128 registers on 16 bits
uint32_t pc;     //16-bit program counter
bool p[3];       //the 3 predicates
bool flags[3];   //processor has 3 flags, V=overflow, N=negative and Z=zero

//DISPLAY DRIVER DEPENDENT
int dispMem[8192]; //should be 4800, but 8129 is 2^13 (there are 13 lines at the
//input of this memory) . This memory is mapped at address:
// 001 0000000000000 to 001 1111111111111
int charROM[256][8];

int mem[8192]; //the memory mapped at address 000 0000000000000 to 000 1111111111111

//------------------------------------------------------------------------------
map<int, coord> regmap;




//------------------------------------------------------------------------------
void printreg(int row, int col, int regnumber, int value){
	//store in a map the screen address where the register is;
	mvprintw(row,col,"REG[%3d]=%04X",regnumber,value);
}

//------------------------------------------------------------------------------

/**
 * converts a binary number under a string format into an integer
 * @param[in] s the input string containing the binary number
 * @return the integer corresponding to the positive binary number contain in s
 */
int b2ToInt(string s)
{
    if (s.length() != 8)
    {
        std::ostringstream o;
        o << "Error in " <<  __FILE__ << "@" << __LINE__ << ": length of s=" << s << " is not 8 ";
        throw runtime_error(o.str());
    }
    unsigned int value=0;
    for (uint8_t i=0;i<s.length();i++)
        value += (1<<(7-i))*(s[i]=='1'?1:0);

    return value;
}

/**
 * Takes as input an unsigned integer x and an selector s.
 * Returns true if in the binary representation of x, the
 * corresponding coefficient of 2^s is 1, else false
 * @param[in] x - the input unsigned number
 * @param[in] s - the coefficient needed to be checked
 * @return 1 if in binary representation coeff. of 2^s is 1, else 0
 */
bool char2bool(unsigned char x, unsigned char s)
{
    unsigned char y;
    y=x>>s;
    return y%2;
}

/**
 * initializes the ROM memory with the corresponding characters
 */
void charROMInit()
{
    ifstream is("rom.data"); // open a file stream

    string line; // S string for a line to read
    int i=0, j=0;
    while (is.good()) // As long as there's data
    {
        is >> line; // read a line
        if (is.good())
        {
            charROM[i][j]=b2ToInt(line.substr(0,8));
            //cout << line.substr(0,8)<<"="<< charROM[i][j] << endl;
            j = (j+1) % 8;
            if (j==0)
                i = i+1;
         }
    }
}

/**
 * parses 1 line from .obj file containing the instructions
 */
void parse(string s) {
	uint16_t adr = 0x0800; //the address where the code segment starts
	ifstream is(s.c_str()); // open a file stream
	string line;             // S string for a line to read
	while(is.good())         // As long as there's data
	{
		is >> line;          // read a line
		istringstream sstr(line);
		sstr >> hex >> mem[adr];
		adr++;
	}
}

/*Change this for memory related display*/
/*we do not have a video memory, but a character memory*/
void render_all()
{
    for (int y = 0; y < 60; y++)
    {
        for (int x = 0; x < 80; x++)
        {
            for (uint8_t k1=0;k1<8;k1++)
                for (uint8_t k2=0;k2<8;k2++)
                {
                    unsigned char c = 255 ; //255=white; color is coded RRRGGBBB on 8 bits
                    c=c*char2bool(charROM[dispMem[y*80+x]][k1],7-k2);
                    video[(8*(y)+k1)*(640)+8*x+k2]=((c&0xe0)<<16) + ((c&0x18)<<11) + ((c&0x7)<<5);
                }
        }
    }
}

bool is_arith_op(uint16_t code)
{
    switch(code)
    {
        case 0:   // ldl
        case 1: // ldh
        case 2:  // mv2reg
        case 3:// mv2acc
        case 4:   // rd
        case 5:   // wr
        case 6:   // rdInc
        case 7:   // wrInc
        case 8:   // rdDec
        case 9:   // wrDec
        case 10:   // jmp
        case 11:   // jmr, add Ri to PC
        case 12:   // jmi
        case 13:   // jmrI
        case 14:   // jsr
        case 15:   // jsri
        case 16:   // rts
        case 17:   // nop
        case 23: //clear
            return false;
        case 18:   // add
        case 19:   // sub
        case 20:   // mul
        case 21:   //cmp
        case 22:   // swap
         case 24: //and
        case 25: //or
        case 26: //xor
        case 27: //not
        case 28: //lsr
        case 29: //lsl
        case 30: //ror
        case 31: //rol
            return true;
        default:
            return false;
    }
}

int compute_sub_ov(uint16_t acc,uint16_t reg)
{
    uint16_t sub=acc+((reg^0xffff)+1);
    //VHDL code: (i_reg(15) xor i_acc(15)) and (not (sub_o(15) xor i_acc(15)));
    // FIXME is this VHDL code ok ?
    
    //return ((reg>>15) ^ (acc>>15)) && !((sub>>15) ^ (acc>>15));
    // FIXME Unsure
    return ((reg>>15) ^ (acc>>15)) && ((sub>>15) ^ (acc>>15));
}

int compute_add_ov(uint16_t acc,uint16_t reg)
{
    uint16_t add=acc+reg;
    //VHDL code: (i_reg(15) xor add_o(15)) and (not (i_reg(15) xor i_acc(15)));
    
    return ((reg>>15) ^ (acc>>15)) && !((add>>15) ^ (acc>>15));
}

void step()
{
    bool doInstr = false;
    bool setFlags = false;
    int tempPC=-1;
    bool tempFlags[3];
    bool signRi;
    bool signImm;
    uint16_t dev;
    int j;
    uint16_t val;

    // FETCH
    uint16_t instr = mem[pc];//fetch the instruction
    //DECODE
    uint16_t instrCode  =  instr >> 11;
    uint16_t instrPred  = (instr >> 8) & 0x7;
    uint16_t instrFlag  = (instr >> 7) & 0x1;
    setFlags = (instrFlag==0?0:1);


    uint16_t reg  =  instr & 0x7f;
    uint32_t imm =   instr & 0xff;

    if(verbose)
    {
	    cout << setfill('0')  << " pc="     << setw(4) << hex << pc 
	                        << " instr=" << setw(4) << hex << instr
	                      << " code=" << setw(2) << hex << instrCode
	                      << " imm=" << setw(2) << hex << imm
	                      << " pred="<< setw(1)<<hex<<instrPred<<endl;
	}

    //predicate execution: check if the instruction with instrCode will be executed
    //or the processor will just jump to the next one.
    //TODO check for corectness
    switch (instrPred)
    {
    case 0: // GT
        if (flags[2]==flags[1] && flags[0]==0)
            doInstr = true;
        break;
    case 1: // GE
        if (flags[2]==flags[1])
            doInstr = true;
        break;
    case 2: // EQ
        if (flags[0]==1)
            doInstr = true;
        break;
    case 3: // NE
        if (flags[0]==false)
            doInstr = true;
        break;
    case 4: // LE
        if (flags[2]!=flags[1] || flags[0]==1) // FIXME: bogdan put && instead of ||
            doInstr = true;
        break;
    case 5: // LT
        if (flags[2]!=flags[1])
            doInstr = true;
        break;
    case 6: // OV
        if (flags[2]==1)
            doInstr = true;
        break;
    case 7: // NC
        doInstr = true;
        break;
    }

    for (int i=0;i<3;i++)
        tempFlags[i] = flags[i];

    if (doInstr==true)
    {
        switch (instrCode)
        {
        case 0:   // ldl
            r[0] = imm;
            printreg( regmap[0].row, regmap[0].col , 0, r[0]);
            break;
        case 1: // ldh
            r[0] = (r[0] & 0xFF) + (imm<<8);
            printreg( regmap[0].row, regmap[0].col , 0, r[0]);
            break;
        case 2:  // mv2reg
            r[reg] = r[0];
            printreg( regmap[reg].row, regmap[reg].col , reg, r[reg]);
            break;
        case 3:// mv2acc
            r[0] = r[reg];
            printreg( regmap[0].row, regmap[0].col , 0, r[0]);
            break;
        case 4:   // rd
            //check device
            dev = r[0]>>13;
            if (dev==0)
                r[reg] = mem[r[0] & 0x1FFF];
            else if (dev==1)
                r[reg] = dispMem[r[0] & 0x1FFF];
            else
            {
                std::ostringstream o;
                o << "Error in " <<  __FILE__ << "@" << __LINE__ << ": Address too large, only memory and DisplayMemory are connected to the proc ";
                throw runtime_error(o.str());
            }
            printreg( regmap[reg].row, regmap[reg].col , reg, r[reg]);
            break;
        case 5:   // wr
            //check device
            dev = r[0]>>13;
            if (dev==0)
                mem[r[0] & 0x1FFF] = r[reg];
            else if (dev==1)
                dispMem[r[0] & 0x1FFF] = r[reg];
            else
            {
                {
                    std::ostringstream o;
                    o << "Error in " <<  __FILE__ << "@" << __LINE__ << ": Address too large, only memory and DisplayMemory are connected to the proc ";
                    throw runtime_error(o.str());
                }
                std::ostringstream o;
                o << "Error in " <<  __FILE__ << "@" << __LINE__ << ": Address too large, only memory and DisplayMemory are connected to the proc ";
                throw runtime_error(o.str());
            }
            break;
        case 6:   // rdInc
            //check device
            dev = r[0]>>13;
            if (dev==0)
                r[reg] = mem[r[0] & 0x1FFF];
            else if (dev==1)
                r[reg] = dispMem[r[0] & 0x1FFF];
            else
            {
                std::ostringstream o;
                o << "Error in " <<  __FILE__ << "@" << __LINE__ << ": Address too large, only memory and DisplayMemory are connected to the proc ";
                throw runtime_error(o.str());
            }
            r[0]++;
            printreg( regmap[0].row, regmap[0].col , 0, r[0]);
		    printreg( regmap[reg].row, regmap[reg].col , reg, r[reg]);
            break;
        case 7:   // wrInc
            //check device
            dev = r[0]>>13;
            if (dev==0)
                mem[r[0] & 0x1FFF] = r[reg];
            else if (dev==1)
                dispMem[r[0] & 0x1FFF] = r[reg];
            else
            {
                std::ostringstream o;
                o << "Error in " <<  __FILE__ << "@" << __LINE__ << ": Address too large, only memory and DisplayMemory are connected to the proc ";
                throw runtime_error(o.str());
            }
            r[0]++;
            break;
        case 8:   // rdDec
            //check device
            dev = r[0]>>13;
            if (dev==0)
                r[reg] = mem[r[0] & 0x1FFF];
            else if (dev==1)
                r[reg] = dispMem[r[0] & 0x1FFF];
            else
            {
                std::ostringstream o;
                o << "Error in " <<  __FILE__ << "@" << __LINE__ << ": Address too large, only memory and DisplayMemory are connected to the proc ";
                throw runtime_error(o.str());
            }
            r[0]--;
            printreg( regmap[0].row, regmap[0].col , 0, r[0]);
			printreg( regmap[reg].row, regmap[reg].col , reg, r[reg]);
            break;
        case 9:   // wrDec
            //check device
            dev = r[0]>>13;
            if (dev==0)
                mem[r[0] & 0x1FFF] = r[reg];
            else if (dev==1)
                dispMem[r[0] & 0x1FFF] = r[reg];
            else
            {
                std::ostringstream o;
                o << "Error in " <<  __FILE__ << "@" << __LINE__ << ": Address too large, only memory and DisplayMemory are connected to the proc ";
                throw runtime_error(o.str());
            }
            r[0]--;
            break;
        case 10:   // jmp
            tempPC = r[reg];
            break;
        case 11:   // jmr, add Ri to PC
            //check sign of R[i]
            signRi = r[reg]>>15;
            if (signRi==0)
                tempPC = pc + r[reg];
            else
                tempPC = pc - r[reg];//TODO
            break;
        case 12:   // jmi
            if (imm>127) //trying to jump at a negative address
            {
                std::ostringstream o;
                o << "Error in " <<  __FILE__ << "@" << __LINE__ << ": Trying to jump with negative immediate "<<imm-256<<"."<<endl;
                throw runtime_error(o.str());;
            }
            tempPC = imm;
            break;
        case 13:   // jmrI,
            //check sign of imm
            signImm = imm>>7;
            if (signImm==0)
                tempPC = pc + imm; //jump forward
            else
                tempPC = pc - ((imm ^ 0xFF)+1); //jump back
            break;
        case 14:   // jsr,
            //TODO check if we are not storing pc in video memory :P
            mem[r[127]]=pc;
            r[127]++;
            tempPC = r[reg];
            printreg( regmap[127].row, regmap[127].col , 127, r[127]);
            break;
        case 15:   // jsri,
            /*
            //check sign of imm
            mem[r[127]] = pc;
            r[127]++;
            if (imm>127) //trying to jump at a negative address
            {
                std::ostringstream o;
                o << "Error in " <<  __FILE__ << "@" << __LINE__ << ": Trying to jump with negative immediate "<<imm-256<<"."<<endl;
                throw runtime_error(o.str());
            }
            tempPC = imm;
            */
            mem[r[127]]=pc;
            r[127]++;
            
            signImm = imm>>7;
            if (signImm==0)
                tempPC = pc + imm; //jump forward
            else
                tempPC = pc - ((imm ^ 0xFF)+1); //jump back
            
            printreg( regmap[127].row, regmap[127].col , 127, r[127]);
            break;
        case 16:   // rts,
            //check sign of imm
            r[127]--; //this is positive (or should be)
            tempPC= mem[r[127]]+1;
            printreg( regmap[127].row, regmap[127].col , 127, r[127]);
            break;
        case 17:   // nop,
            break;
        case 18:   // add,
            {
            flags[2]=compute_add_ov(r[0],r[reg]);
            
            r[0] +=r[reg];
            flags[0]=(r[0]==0);
            flags[1]=((r[0]>>15)==1);
            }
            printreg( regmap[0].row, regmap[0].col , 0, r[0]);
            break;
        case 19:   // sub,
            {
            flags[2]=compute_sub_ov(r[0],r[reg]);
            
            r[0] -=r[reg];
            flags[0]=(r[0]==0);
            flags[1]=((r[0]>>15)==1);
            }
            printreg( regmap[0].row, regmap[0].col , 0, r[0]);
            break;
        case 20:   // mul,
            {
            bool neg=false;
            int res;
            uint16_t r0=r[0];
            uint16_t r1=r[reg];
            uint16_t ans;
            if((r0>>15))
            {
                r0=(r0^0xffff)+1;
                neg=!neg;
            }
            
            if((r1>>15))
            {
                r1=(r1^0xffff)+1;
                neg=!neg;
            }
            r0 = r0 & 0xff;
            r1 = r1 & 0xff;
            ans=r0*r1;
            if(neg)
                ans=(ans^0xffff)+1;
            r[0]=ans;
            res=int(ans);
            //flags[0]=(res==0);
            //flags[1]=(res<0);
            //flags[2]=0;
            /*
            int sign1=(r[0]>>15 ? -1 : 1);
            int v0=(sign1*r[0])&0xFF;
            int sign2=(r[reg]>>15 ? -1 : 1);
            int v1=(sign2*r[reg])&0xFF;
            
            r[0]=sign1*sign2*v0*v1;
            */
            printreg( regmap[0].row, regmap[0].col , 0, r[0]);
            }
            break;
        case 21:   //cmp
            {
            uint16_t temp=r[0];
            temp+=(r[reg]^0xffff)+1;
            flags[0]=(temp==0);
            flags[1]=((temp>>15)==1);
            flags[2]=compute_sub_ov(r[0],r[reg]);
            }
            break;
        case 22:   // swap,
            r[reg] = (r[reg] >> 8) + ((r[reg] & 0xFF)<<8);
            printreg( regmap[reg].row, regmap[reg].col , reg, r[reg]);
            break;
        case 23: //clear
            r[reg]=0;
            printreg( regmap[reg].row, regmap[reg].col , reg, r[reg]);
            break;
        case 24: //and
            r[0] = r[0] & r[reg];
            printreg( regmap[0].row, regmap[0].col , 0, r[0]);
            break;
        case 25: //or
            r[0] = r[0] | r[reg];
            printreg( regmap[0].row, regmap[0].col , 0, r[0]);
            break;
        case 26: //xor
            r[0] = r[0] ^ r[reg];
            printreg( regmap[0].row, regmap[0].col , 0, r[0]);
            break;
        case 27: //not
            r[0] = r[reg] ^ 0xFFFF;
            printreg( regmap[0].row, regmap[0].col , 0, r[0]);
            break;
        case 28: //lsr
            r[0] = (r[0]>>r[reg]);
            printreg( regmap[0].row, regmap[0].col , 0, r[0]);
            break;
        case 29: //lsl
            r[0] = (r[0]<<r[reg]);
            printreg( regmap[0].row, regmap[0].col , 0, r[0]);
            break;
        case 30: //ror
            // FIXME: amaury: c'est faux si r[reg]>15 ?
            val=0;
            for (j=0;j<r[reg];j++)
                val += (2<<j);
            r[0] = (r[0]>>r[reg]) + ((r[0] & val)<<r[reg]);
            printreg( regmap[0].row, regmap[0].col , 0, r[0]);
            break;
        case 31: //rol
            // FIXME: amaury: c'est faux si r[reg]>15 ?
            val=0;
            for (j=15;j>15-r[reg];j--)
                val+= (2<<j);
            r[0] = (r[0]>>r[reg]) + ((r[0]&val)>>(16-r[reg])); //TODO
            printreg( regmap[0].row, regmap[0].col , 0, r[0]);
            break;
        }
    }
    
    if(doInstr && instrCode!=21)
    {
        if(r[0]==0)
            flags[0]=1;
        else
            flags[0]=0;
    
        if(is_arith_op(instrCode))
        {
            if((r[0]>>15)==1)
                flags[1]=1;
            else
                flags[1]=0;
        }
        
        if(instrCode!=18 && instrCode!=19 && instrCode!=20)
            flags[2]=0;// TODO
    }

    if (doInstr)
        if (tempPC>-1)
            pc = tempPC;
        else
            pc++;
    else
    {
        pc++;
    }
    
    // TODO: check PC
    if(pc>8192)
    {
        std::ostringstream o;
        o << "Error in " <<  __FILE__ << "@" << __LINE__ << ": PC ("<<hex<<pc<<") is out of range !";
        throw runtime_error(o.str());
    }

    if (!setFlags)
        for (j=0;j<3;j++)
            flags[j]=tempFlags[j];

    int row, col;
	getmaxyx(stdscr,row,col);
	mvprintw(1,col-9,"O = %d",flags[2]);
	mvprintw(2,col-9,"N = %d",flags[1]);
	mvprintw(3,col-9,"Z = %d",flags[0]);
	mvprintw(4,col-15,"DoInstr = %d",doInstr);
	mvprintw(4,col-15,"Instr = %04X",instr);
	string instr_str=disassemble(instr);
	mvprintw(5,col-18,"               ");
	mvprintw(5,col-instr_str.size()-1,"%s",instr_str.c_str());
}

bool infinite_loop()
{
    uint16_t instr = mem[pc];//fetch the instruction
    //DECODE
    uint16_t instrCode  =  instr >> 11;
    uint16_t instrPred  = (instr >> 8) & 0x7;

    uint32_t imm =   instr & 0xff;
    
    if(instrCode==13 && instrPred==7 && imm==0)
        return true;
    else
        return false;
}

void printregs()
{
	int row,col;
	getmaxyx(stdscr,row,col);
	//number of registers 
	int reg_columns = col / 14;
	int real_columns = (NREGS+row-1)/row;

	if (reg_columns*row < NREGS){	
		cerr << "Cannot display all registers" <<endl;
		sleep(10);
		exit(1);
	}

	for (int i=0; i<reg_columns; i++)
		for (int j=0; j<row; j++)
			if (i*row+j<NREGS)
			{
				coord c;
				c.row = j;
				c.col = i*14;
				regmap[i*row+j]=c;
				printreg(j,i*14,i*row+j,0);
			}	
	for (int j=0; j<row; j++)
		mvprintw(j,14*real_columns+1,"|");
}

bool g_quit=false;
int g_result=0;
int reg_ret=-1;
bool g_quit_on_loop=false;
bool g_nox=false;

void CtrlC(int)
{
    if(g_quit)
        exit(g_result);
    endwin();
    g_quit=true;
}

int _main(int argc,char* argv[])
{
    verbose = 0;
    step_by_step = 0;
    string filename;

    if(argc<=1)
    {
        cout<<"Usage: "<<argv[0]<<" [options] <filename>\n";
        cout<<"Options:\n";
        cout<<"  -v\tVerbose\n";
        cout<<"  -s\tStep by step\n";
        cout<<"  -r<n>\tReturn the content of r<n> instead of 0 upon exit\n";
        cout<<"  -ql\tQuit on infinite loop\n";
        cout<<"  -nox\tNo X window\n";
        
        return EXIT_FAILURE;
    }

    if(argc>=3)
    {
        for(int i=1;i<argc-1;i++)
        {
            string opt(argv[i]);
            if(opt=="-v")
                verbose=1;
            else if(opt=="-s")
                step_by_step=1;
            else if(opt.size()>2 && opt[0]=='-' && opt[1]=='r')
            {
                istringstream iss(opt.substr(2));
                int reg;
                iss>>reg;
                if(iss.fail())
                {
                    cout<<"invalid command line argument\n";
                    return EXIT_FAILURE;
                }
                reg_ret=reg;
                cout<<"Return reg: "<<reg_ret<<"\n";
            }
            else if(opt=="-ql")
            {
                g_quit_on_loop=true;
                cout<<"Will quit on infinite loop\n";
            }
            else if(opt=="-nox")
            {
                g_nox=true;
                cout<<"Will not display an X window\n";
            }
            else
            {
                cout<<"Invalid command line argument\n";
                return EXIT_FAILURE;
            }
        }
        filename = argv[argc-1];
    }
    else
    filename = argv[1];

    cout << "The filename is " << filename << endl;

    try
    {
        charROMInit();
    }
    catch (std::string s)
    {
        cerr << "Exception while parsing command line: " << s << endl;
        return 1;
    }

    if(step_by_step==1)
        SKIP_FRAME=0;
    
    Display *dpy=0;
    Window w=0;
    XImage *img=0;
    GC gc=0;
    if(!g_nox)
    {
        //X window initialization
        dpy = XOpenDisplay(NULL);
        Atom wm_delete_window = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
        w = XCreateSimpleWindow(dpy, DefaultRootWindow(dpy), 
                     0, 0, 640, 480,
                     0, BlackPixel(dpy, DefaultScreen(dpy)), 
        BlackPixel(dpy, DefaultScreen(dpy)));
        XSetWMProtocols(dpy,w,&wm_delete_window,1);
        XSelectInput(dpy, w, ExposureMask | KeyPressMask | StructureNotifyMask);
        XMapWindow(dpy, w);
        XStoreName(dpy, w, "ASR1 2008 RISC simulator");

        gc = XCreateGC(dpy, w, 0, NULL);

        img = XCreateImage(dpy, DefaultVisual(dpy, DefaultScreen(dpy)), 
                     24, ZPixmap, 0,
                    (char *)video, 640, 480, 32, 640*4);
     }

    // first update to the video memory
    parse(filename);

    initscr();/* Start curses mode   */
    printregs();

    render_all();

    if(!g_nox)
    {
        XPutImage(dpy, w, gc, img, 0, 0, 0, 0, 640, 480);
        XFlush(dpy);
    }

    pc = 0x0800;

    int row,col;
    uint32_t frames = 0;
    time_t start, current;
    time (&start); 
    int fps = 0;
    bool destroyed=false;

    signal(SIGINT,CtrlC);

    int frame=0;
    while(!g_quit)
    {
        if(step_by_step==1)
            getch();
        //compute frames per second:
        //----------------------------------------------------------------------
        getmaxyx(stdscr,row,col);
        time(&current);
        double secdiff = difftime(current,start);
        frames++;  
        if (secdiff > 1)
        {
            fps = int(double(frames) / secdiff);
            time (&start);
            frames = 0;
        }
        mvprintw(row-1,col-6,"%6d",fps);
        //----------------------------------------------------------------------

        refresh();/* Print it on to the real screen */

        //let's print the PC
        mvprintw(0,col-10,"PC = %04X",pc);

        step();

        if(frame==SKIP_FRAME)
        {
            if(!g_nox)
                render_all();

            if(!destroyed && !g_nox)
            {
                XPutImage(dpy, w, gc, img, 0, 0, 0, 0, 640, 480);
                XFlush(dpy);
            }
            
            frame=0;
            
            if(!g_nox && XPending(dpy)>0)
            {
                XEvent e;
                XNextEvent(dpy,&e);
                if(e.type==DestroyNotify)
                {
                    g_quit=true;
                }
                else if(e.type==ClientMessage)
                {
                    if(e.xclient.data.l[0] && e.xclient.window==w)
                    {
                        XDestroyWindow(dpy,w);
                        destroyed=true;
                    }
                }
            }
        }
        else
            frame++;
            
        if(infinite_loop() && g_quit_on_loop)
            g_quit=true;
            
        if(reg_ret>=0 && reg_ret<NREGS)
            g_result=r[reg_ret];
    }
    
    if(!destroyed && !g_nox)
        XDestroyWindow(dpy,w);

    endwin();/* End curses mode  */
    if(!g_nox)
        XCloseDisplay(dpy);

    cout<<"Exiting normally...\n";

    return g_result;
}

int main(int argc,char **argv)
{
    try
    {
        return _main(argc,argv);
    }
    catch(exception& e)
    {
        endwin();
        cout<<"An exception occured: "<<e.what()<<"\n";
    }
    catch(...)
    {
        endwin();
        cerr<<"An unknown exception occured\n";
    }
    
    return -1;
}
